#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "sosemanuk.h"

#define		PRNG_DEV_PRIMARY			"/dev/urandom"
#define		PRNG_DEV_SECONDARY			"/dev/random" /* Too slow on most *nix */	
#define		DEFAULT_SAMPLE_SIZE			1024
#define		DEFAULT_POOL_SIZE			1024
#define		RAINMAKER_SESSION_KEY_SIZE		16
#define		RAINMAKER_IV_SIZE			16

#define		RAINMAKER_BANNER		"Rainmaker Pseudo Random Number Generator v1.0\nWritten by Frank Gardner\n\n"

#define		rainmaker_only_print(x)		isprint(x)
#define		rainmaker_only_alpha(x)		isalpha(x)
#define		rainmaker_only_num(x)		isdigit(x)
#define		rainmaker_only_alnum(x)		isalpha(x)

typedef unsigned char 	u8;
typedef unsigned long 	u32;
typedef unsigned short 	u16;

u8 rainmaker_entropy_pool[DEFAULT_POOL_SIZE + 1], rainmaker_key[RAINMAKER_SESSION_KEY_SIZE];	
int random_fd;
sosemanuk_key_context ctx;
sosemanuk_run_context rc;


__inline__ void rainmaker_encrypt(u8 *, u32);

/* Rainmaker Structure */
typedef struct
{
	int 	hexmode;
	int 	o_a, o_an, o_n, o_p;
	u32	size;
} rainmaker_t;


__inline__ void 
rainmaker_print_hex(u8 x)
{
	printf("%02X",x);	

}


__inline__ void rainmaker_init(void)
{
	
	if((random_fd = open(PRNG_DEV_PRIMARY, O_RDONLY)) < 0 )
	{
		if((random_fd = open(PRNG_DEV_SECONDARY, O_RDONLY)) < 0)
		{
			fprintf(stderr,"rainmaker: cannot find a reliable PRNG device on this host.\n");
			exit(-1);
		}
	}


	while((read(random_fd, rainmaker_entropy_pool, DEFAULT_POOL_SIZE)) < DEFAULT_POOL_SIZE)
	{
		continue;
	}


	while((read(random_fd, rainmaker_key, RAINMAKER_SESSION_KEY_SIZE)) < RAINMAKER_SESSION_KEY_SIZE)
	{
		continue;

	}
	
	sosemanuk_schedule(&ctx, rainmaker_key, 16);		
 
}

__inline__ void rainmaker_fill_random_pool(void)
{
	static int x = 0;


	/* Refill the entropy pool every three runs */ 
	if(!x)
	{
       		 while((read(random_fd, rainmaker_entropy_pool, DEFAULT_POOL_SIZE)) < DEFAULT_POOL_SIZE)
       		 {
                	continue;
        	 }
		 x = 3;
	}
	
	rainmaker_encrypt(rainmaker_entropy_pool, DEFAULT_POOL_SIZE);
	--x;
}


__inline__ void rainmaker_encrypt(u8 *x, u32 y)
{

	u8 iv[RAINMAKER_IV_SIZE];

	while((read(random_fd, iv, RAINMAKER_IV_SIZE)) < RAINMAKER_IV_SIZE)
	{
		continue;
	}

	sosemanuk_init(&rc, &ctx, iv, RAINMAKER_IV_SIZE);
	sosemanuk_encrypt(&rc, x, x, y);
	
}


__inline__ void rainmaker_main(rainmaker_t *rm)
{

	register u32 x = 0, pool_size = 0;


	x = rm->size;

	while(x)
	{
		if(!pool_size)
	
		{
			pool_size = DEFAULT_POOL_SIZE;
			rainmaker_fill_random_pool();
		}
		
		if(rm->o_an)
		{
			if(rainmaker_only_alnum(rainmaker_entropy_pool[pool_size-1]) == 0)
			{
				pool_size--;
				continue;
			}	

		}
		else
		if(rm->o_a)
		{
			if(rainmaker_only_alpha(rainmaker_entropy_pool[pool_size-1]) == 0)
			{
				pool_size--;
				continue;
			}	

		}
		else
		if(rm->o_n)
		{
			if(rainmaker_only_num(rainmaker_entropy_pool[pool_size-1]) == 0)
			{
				pool_size--;
				continue;
			}	

		}
		else
		if(rm->o_p)
		{
			if(rainmaker_only_print(rainmaker_entropy_pool[pool_size-1]) == 0)
			{
				pool_size--;
				continue;
			}	

		}

		if(rm->hexmode)
		{
			rainmaker_print_hex(rainmaker_entropy_pool[pool_size-1]);
			pool_size--;
			x--;
		}
		else
		{
			putchar(rainmaker_entropy_pool[pool_size-1]);
			pool_size--;
			x--;
		}

		
		
	}

	fflush(stdout);
}


int main(int argc, char **argv)
{
	register int c = 0, j = 0;
	rainmaker_t rm;

	
	bzero(&rm,sizeof(rm));

	j = rm.size = DEFAULT_POOL_SIZE;
	
	
	while((c = getopt(argc,argv,"anpHhs:")) != -1)
	{

		switch(c)
		{
			case 'a':
					rm.o_a = 1;
					break;
			case 'n':
					rm.o_n = 1;
					break;
			case 'p':
					rm.o_p = 1;
					break;
			case 'H':
					rm.hexmode = 1;
					break;
			case 'h':
			                printf("%s",RAINMAKER_BANNER);
					printf("Usage: %s -[anpHhs]\n",argv[0]);
					printf("-a\tPrint only alphabetic characters\n");
					printf("-n\tPrint only numeric characters\n");
					printf("-p\tPrint only printable characters\n");
					printf("-H\tPrint characters in Base16 format (Hexidecimal)\n");
					printf("-h\tThis help\n");
					printf("-s\tAmount of random data to generate (default: %d bytes)\n\n",DEFAULT_POOL_SIZE);
                			return 0;

			case 's':
					j = atoi(optarg);
					rm.size = j;
					break;
		}

	}


	if(j <=0)
	{
		fprintf(stderr,"Incorrect value suppied to '-s'\n");
		return -1;
	}
	rainmaker_init();
	rainmaker_main(&rm);

	return 0;
}
